@@ -3,7 +3,7 @@
 
 
 #define PACKAGE cscope
-#define VERSION 15.1
+#define VERSION 15.4
 
 /* OS Definitions */
 #undef Linux
@@ -36,4 +36,8 @@
  *   2 = version 4.0/4.1 */
 #undef NCURSES_970530
 
+/* Define this if the lex used is the 'real' AT&T variety. Don't define if
+ * it's flex or some other */
+#undef USING_LEX
+
 
